# distutils: language = c++
# distutils: include_dirs = dwave/samplers/tabu/src/
# distutils: sources = dwave/samplers/tabu/src/tabu_search.cpp dwave/samplers/tabu/src/tabu_utils.cpp dwave/samplers/tabu/src/bqp.cpp

# Copyright 2020 D-Wave Systems Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from libcpp.vector cimport vector
from libc.time cimport time
import numpy as np

cimport dwave.samplers.tabu.tabu


cdef class TabuSearch:
    """Wraps the class `TabuSearch` from `src/tabu_search.cpp`."""

    cdef dwave.samplers.tabu.tabu.TabuSearch *c_tabu

    def __cinit__(self,
                  object Q,
                  object initSol,
                  int tenure,
                  int timeout,
                  int numRestarts,
                  object seed=None,
                  object energyThreshold=None,
                  object coeffZFirst=None,
                  object coeffZRestart=None,
                  object lowerBoundZ=None):
        cdef unsigned int _seed = time(NULL) if seed is None else seed
        cdef double _energyThreshold = -np.inf if energyThreshold is None else energyThreshold
        cdef int _coeffZFirst = -1 if coeffZFirst is None else coeffZFirst
        cdef int _coeffZRestart = -1 if coeffZRestart is None else coeffZRestart
        cdef int _lowerBoundZ = -1 if lowerBoundZ is None else lowerBoundZ

        cdef double[:,:] qubo = np.asarray(Q, dtype=np.double)
        cdef vector[vector[double]] Qvec
        Qvec.resize(qubo.shape[0])
        cdef Py_ssize_t i, j
        for i in range(qubo.shape[0]):
            for j in range(qubo.shape[1]):
                Qvec[i].push_back(qubo[i, j])

        cdef int[:] initial = np.asarray(initSol, dtype=np.intc)
        cdef vector[int] initVec
        for i in range(len(initial)):
            initVec.push_back(initial[i])

        with nogil:
            self.c_tabu = new dwave.samplers.tabu.tabu.TabuSearch(
                Qvec, initVec, tenure, timeout, numRestarts, _seed, _energyThreshold,
                _coeffZFirst, _coeffZRestart, _lowerBoundZ)

    def __dealloc__(self):
        del self.c_tabu

    def bestEnergy(self):
        return self.c_tabu.bestEnergy()

    def bestSolution(self):
        return self.c_tabu.bestSolution()

    def numRestarts(self):
        return self.c_tabu.numRestarts()



# distutils: language = c++
# cython: language_level = 3
# distutils: include_dirs = dwave/samplers/tree/src/include/
#
# Copyright 2019 D-Wave Systems Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from typing import Tuple
from libcpp cimport bool
from libc.stdlib cimport free

import numpy as np
import dimod
from dimod.binary.binary_quadratic_model import BinaryQuadraticModel

cimport numpy as np
from dimod cimport cyBQM_float64
from dimod.libcpp cimport cppBinaryQuadraticModel

from dwave.samplers.tree.orang cimport samples_type, PyArray_ENABLEFLAGS

cdef extern from "src/include/sample.hpp":
    void sampleBQM[V, B](cppBinaryQuadraticModel[B, V]& refBQM,
                         int* var_order,
                         double beta,
                         int low,
                         double max_complexity,
                         int num_samples,
                         bool marginals,
                         int seed,
                         double* log_pf,
                         int** samples_data, int* samples_rows, int* samples_cols,
                         double** single_mrg_data, int* single_mrg_len,
                         double** pair_mrg_data, int* pair_mrg_rows, int* pair_mrg_cols,
                         int** pair_data, int* pair_rows, int* pair_cols) except +

def sample_bqm_wrapper(bqm: BinaryQuadraticModel,
                       beta: float,
                       max_complexity: float,
                       order: list,
                       marginals: bool = False,
                       num_reads: int = 1,
                       seed: float = None) -> Tuple[np.ndarray, dict]:
    """Cython wrapper for :func:`sampleBQM`.

    Args:
        bqm:
            Binary quadratic model to sample from. Variables must be linearly indexed.

        beta:
            `Boltzmann distribution`_ inverse temperature parameter.

        max_complexity:
            Upper bound on algorithm's complexity.

        order:
            List of variables representing the variable elimination order.

        marginals:
            Determines whether or not to compute the marginals. If True, they
            will be included in the returned dict.

        num_reads:
            Number of samples to return.

        seed:
            Random number generator seed. Negative values will cause a time-based
            seed to be used.

    Returns:
        The samples and marginals.
    """
    if not bqm.num_variables:
        raise ValueError("bqm must have at least one variable.")

    if len(order) != bqm.num_variables or set(order) != bqm.variables:
        raise ValueError("order must contain the variables in bqm.")

    cdef cyBQM_float64 cybqm = bqm.data

    cdef double _beta = beta
    cdef double _max_complexity = max_complexity
    cdef int low = -1 if bqm.vartype is dimod.SPIN else 0

    cdef int _seed
    if seed is None:
        _seed = np.random.randint(np.iinfo(np.intc).max, dtype=np.intc)
    else:
        _seed = seed

    cdef int[:] elimination_order = np.asarray(order, dtype=np.intc)
    cdef int* elimination_order_ptr = &elimination_order[0]

    # pass in pointers so that sample_bqm can fill things in
    cdef double logpf

    # samples
    cdef int srows, scols
    cdef samples_type* samples_pointer

    # marginals
    cdef double* single_marginals_pointer
    cdef int smlen

    cdef double* pair_marginals_pointer
    cdef int pmrows, pmcols

    cdef int* pair_pointer
    cdef int prows, pcols

    sampleBQM(cybqm.cppbqm,
              elimination_order_ptr,
              _beta,
              low,
              _max_complexity,
              num_reads,
              marginals,
              _seed,
              &logpf,
              &samples_pointer, &srows, &scols,
              &single_marginals_pointer, &smlen,
              &pair_marginals_pointer, &pmrows, &pmcols,
              &pair_pointer, &prows, &pcols)

    # create a numpy array without making a copy then tell numpy it needs to
    # free the memory
    samples = np.asarray(<samples_type[:srows, :scols]> samples_pointer)
    PyArray_ENABLEFLAGS(samples, np.NPY_OWNDATA)

    # convert the samples to spin if necessary
    cdef size_t i, j
    if low == -1:
        for i in range(srows):
            for j in range(scols):
                if samples[i, j] == 0:
                    samples[i, j] = -1

    if marginals:
        variable_marginals = np.asarray(<double[:smlen]> single_marginals_pointer)
        PyArray_ENABLEFLAGS(variable_marginals, np.NPY_OWNDATA)

        if pmrows * pmcols:
            interaction_marginals = np.asarray(<double[:pmrows, :pmcols]> pair_marginals_pointer)
            PyArray_ENABLEFLAGS(interaction_marginals, np.NPY_OWNDATA)
        else:
            interaction_marginals = np.empty(shape=(pmrows, pmcols), dtype=np.double)

        if prows * pcols:
            interactions = np.asarray(<int[:prows, :pcols]> pair_pointer)
            PyArray_ENABLEFLAGS(interactions, np.NPY_OWNDATA)
        else:
            interactions = np.empty(shape=(prows, pcols), dtype=np.double)

        marginal_data = dict(variable_marginals=variable_marginals,
                             interactions=interactions,
                             interaction_marginals=interaction_marginals,
                             log_partition_function=logpf,
                             )
    else:
        marginal_data = dict(log_partition_function=logpf)

    marginal_data['log_partition_function'] += -beta*bqm.offset

    return samples, marginal_data



# distutils: language = c++
# distutils: include_dirs = dwave/samplers/sa/src/
# distutils: sources = dwave/samplers/sa/src/cpu_sa.cpp
# cython: language_level = 3

# Copyright 2018 D-Wave Systems Inc.
#
#    Licensed under the Apache License, Version 2.0 (the "License");
#    you may not use this file except in compliance with the License.
#    You may obtain a copy of the License at
#
#        http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS,
#    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#    See the License for the specific language governing permissions and
#    limitations under the License.

from libcpp cimport bool
from libcpp.vector cimport vector

import numpy as np
cimport numpy as np


cdef extern from "cpu_sa.h":
    ctypedef bool (*callback)(void *function)

    int general_simulated_annealing(
            char* samples,
            double* energies,
            const int num_samples,
            const vector[double] & h,
            const vector[int] & coupler_starts,
            const vector[int] & coupler_ends,
            const vector[double] & coupler_weights,
            const int sweeps_per_beta,
            const vector[double] & beta_schedule,
            const unsigned long long seed,
            callback interrupt_callback,
            void *interrupt_function) nogil


def simulated_annealing(num_samples, h, coupler_starts, coupler_ends,
                        coupler_weights, sweeps_per_beta, beta_schedule, seed,
                        np.ndarray[char, ndim=2, mode="c"] states_numpy,
                        interrupt_function=None):
    """Wraps `general_simulated_annealing` from `cpu_sa.cpp`. Accepts
    an Ising problem defined on a general graph and returns samples
    using simulated annealing.

    Parameters
    ----------
    num_samples : int
        Number of samples to get from the sampler.

    h : list(float)
        The h or field values for the problem.

    coupler_starts : list(int)
        A list of the start variable of each coupler. For a problem
        with the couplers (0, 1), (1, 2), and (3, 1), `coupler_starts`
        should be [0, 1, 3].

    coupler_ends : list(int)
        A list of the end variable of each coupler. For a problem
        with the couplers (0, 1), (1, 2), and (3, 1), `coupler_ends`
        should be [1, 2, 1].

    coupler_weights : list(float)
        A list of the J values or weight on each coupler, in the same
        order as `coupler_starts` and `coupler_ends`.

    sweeps_per_beta : int
        The number of sweeps to perform at each beta value provided in
        `beta_schedule`. The total number of sweeps per sample is
        sweeps_per_beta * len(beta_schedule).

    beta_schedule : list(float)
        A list of the different beta values to run sweeps at.

    seed : 64 bit int > 0
        The seed to use for the PRNG. Must be a positive integer
        greater than 0. If the same seed is used and the rest of the
        parameters are the same, the returned samples will be
        identical.

    states_numpy : np.ndarray[char, ndim=2, mode="c"], values in (-1, 1)
        The initial seeded states of the simulated annealing runs. Should be of
        a contiguous numpy.ndarray of shape (num_samples, num_variables).

    interrupt_function: function
        Should accept no arguments and return a bool. The function is
        called between samples and if it returns True, simulated annealing
        will return early with the samples it already has.

    Returns
    -------
    samples : numpy.ndarray
        A 2D numpy array where each row is a sample.

    energies: np.ndarray
        The energies.

    """
    num_vars = len(h)

    # in the case that we either need no samples or there are no variables,
    # we can safely return an empty array (and set energies to 0)
    if num_samples*num_vars == 0:
        annealed_states = np.empty((num_samples, num_vars), dtype=np.int8)
        return annealed_states, np.zeros(num_samples, dtype=np.double)

    # allocate ndarray for energies
    energies_numpy = np.empty(num_samples, dtype=np.float64)
    cdef double[:] energies = energies_numpy

    # explicitly convert all Python types to C while we have the GIL
    cdef char* _states = &states_numpy[0, 0]
    cdef double* _energies = &energies[0]
    cdef int _num_samples = num_samples
    cdef vector[double] _h = h
    cdef vector[int] _coupler_starts = coupler_starts
    cdef vector[int] _coupler_ends = coupler_ends
    cdef vector[double] _coupler_weights = coupler_weights
    cdef int _sweeps_per_beta = sweeps_per_beta
    cdef vector[double] _beta_schedule = beta_schedule
    cdef unsigned long long _seed = seed

    cdef void* _interrupt_function
    if interrupt_function is None:
        _interrupt_function = NULL
    else:
        _interrupt_function = <void *>interrupt_function

    with nogil:
        num = general_simulated_annealing(_states,
                                          _energies,
                                          _num_samples,
                                          _h,
                                          _coupler_starts,
                                          _coupler_ends,
                                          _coupler_weights,
                                          _sweeps_per_beta,
                                          _beta_schedule,
                                          _seed,
                                          interrupt_callback,
                                          _interrupt_function)

    # discard the noise if we were interrupted
    return states_numpy[:num], energies_numpy[:num]


cdef bool interrupt_callback(void * const interrupt_function) with gil:
    try:
        return (<object>interrupt_function)()
    except Exception:
        # if an exception occurs, treat as an interrupt
        return True



# Copyright 2019 D-Wave Systems Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from typing import List, Optional

import dimod
import numpy as np

from dimod.typing import Variable

from dwave.samplers.tree.sample import sample_bqm_wrapper
from dwave.samplers.tree.solve import solve_bqm_wrapper, samples_dtype, energies_dtype
from dwave.samplers.tree.utilities import elimination_order_width, min_fill_heuristic

__all__ = ['TreeDecompositionSolver', 'TreeDecompositionSampler']


class TreeDecompositionSolver(dimod.Sampler):
    """Tree decomposition-based solver for binary quadratic models.

    The tree decomposition solver uses `tree decomposition`_ to find ground
    states of the given binary quadratic model.

    .. Important:: The performance of this sampler is highly dependant on the
        treewidth of the problem.

    Examples:
        Create a solver:

        >>> from dwave.samplers import TreeDecompositionSolver
        >>> solver = TreeDecompositionSolver()

        Create a simple Ising problem:

        >>> h = {'a': .1, 'b': 0}
        >>> J = {('a', 'b') : -1}

        Find the ground state.

        >>> sampleset = solver.sample_ising(h, J)
        >>> sampleset.first.sample
        {'a': -1, 'b': -1}

        Take multiple reads to find states of increasing energy.

        >>> sampleset = solver.sample_ising(h, J, num_reads=3)
        >>> print(sampleset)
           a  b energy num_oc.
        0 -1 -1   -1.1       1
        1 +1 +1   -0.9       1
        2 -1 +1    0.9       1
        ['SPIN', 3 rows, 3 samples, 2 variables]

    .. _tree decomposition: https://en.wikipedia.org/wiki/Tree_decomposition

    """
    parameters = {'num_reads': [],
                  'elimination_order': ['max_treewidth']}
    """Keyword arguments accepted by the sampling methods.

    Examples:

        >>> from dwave.samplers import TreeDecompositionSolver
        >>> solver = TreeDecompositionSolver()
        >>> solver.parameters.keys()
        dict_keys(['num_reads', 'elimination_order'])

    See :meth:`.sample` for descriptions.

    """

    properties = {'max_treewidth': 25}
    """Information about the solver.

    Examples:

        >>> from dwave.samplers import TreeDecompositionSolver
        >>> solver = TreeDecompositionSolver()
        >>> solver.properties
        {'max_treewidth': 25}

    """

    def __init__(self):
        # make these object properties rather than class properties
        self.parameters = dict(self.parameters)
        self.properties = dict(self.properties)

    def sample(self, bqm: dimod.BinaryQuadraticModel, num_reads: Optional[int] = 1,
               elimination_order: Optional[List[Variable]] = None) -> dimod.SampleSet:
        """Find ground states of a binary quadratic model.

        Args:
            bqm:
                Binary quadratic model (BQM).

            num_reads:
                Total number of samples to draw. The samples are drawn in
                order of energy so if ``num_reads=1``, only the ground state is
                returned. If ``num_reads=2``, the ground state and the first
                excited state are returned. If ``num_reads >= len(bqm)**2``,
                samples are duplicated.

            elimination_order:
                Variable elimination order. Should be a list of the
                variables in the binary quadratic model. If None, the min-fill
                heuristic [#gd]_ is used to generate one.

        Raises:
            ValueError:
                The treewidth_ of the given BQM and elimination order cannot
                exceed the value provided in :attr:`.properties`.

        .. _treewidth: https://en.wikipedia.org/wiki/Treewidth

        .. [#gd] Gogate & Dechter, "A Complete Anytime Algorithm for Treewidth",
           https://arxiv.org/abs/1207.4109

        """
        if not bqm:
            samples = np.empty((num_reads, 0), dtype=samples_dtype)
            energies = bqm.energies(samples, dtype=energies_dtype)
            return dimod.SampleSet.from_samples(samples, bqm.vartype, energy=energies)

        bqm = dimod.as_bqm(bqm, copy=True, dtype=float)

        max_samples = min(num_reads, 2**len(bqm))

        if elimination_order is None:
            tree_width, elimination_order = min_fill_heuristic(bqm)
        else:
            tree_width = elimination_order_width(bqm, elimination_order)

        # developer note: we start getting bad_alloc errors above tree_width 25, this
        # should be fixed in the future
        if tree_width > self.properties['max_treewidth']:
            raise ValueError(
                f"maximum treewidth of {self.properties['max_treewidth']} exceeded. "
                "To see the bqm's treewidth:\n"
                ">>> import dwave_networkx as dnx\n"
                f">>> dnx.elimination_order_width(bqm.adj, {elimination_order})"
                )

        max_complexity = tree_width + 1

        # relabel bqm variables so that we only work with linear indices
        bqm_copy, int_to_var = bqm.relabel_variables_as_integers(inplace=False)
        var_to_int = {v: k for k, v in int_to_var.items()}

        # relabel variables in the elimination order as well
        elimination_order = [var_to_int.get(var, var) for var in elimination_order]

        samples, energies = solve_bqm_wrapper(bqm=bqm_copy,
                                              order=elimination_order,
                                              max_complexity=max_complexity,
                                              max_solutions=max_samples
                                              )

        # if we asked for more than the total number of distinct samples, we
        # just resample again starting from the beginning
        num_occurrences = np.ones(max_samples, dtype=np.intc)
        if num_reads > max_samples:
            q, r = divmod(num_reads, max_samples)
            num_occurrences *= q
            num_occurrences[:r] += 1

        return dimod.SampleSet.from_samples((samples, bqm.variables),
                                            bqm.vartype,
                                            energy=energies,
                                            num_occurrences=num_occurrences)


class TreeDecompositionSampler(dimod.Sampler):
    """Tree decomposition-based sampler for binary quadratic models.

    The tree decomposition sampler uses `tree decomposition`_ to sample from a
    `Boltzmann distribution`_ defined by the given binary quadratic model.

    .. Important:: The performance of this sampler is highly dependant on the
        treewidth of the problem.

    Examples:
        Create a sampler:

        >>> from dwave.samplers import TreeDecompositionSampler
        >>> sampler = TreeDecompositionSampler()

        Create a simple Ising problem:

        >>> h = {'a': .1, 'b': 0}
        >>> J = {('a', 'b') : -1}

        Sample from the given problem.

        >>> sampleset = sampler.sample_ising(h, J, num_reads=100,
        ...                                  elimination_order=['a', 'b'])
        >>> sampleset.first.sample
        {'a': -1, 'b': -1}

        View information about the distribution.

        >>> variable_marginals = sampleset.info['variable_marginals']
        >>> round(variable_marginals['a'], 3)  # prob(a == 1)
        0.354
        >>> round(1 - variable_marginals['b'], 3)  # prob(b == -1)
        0.645

        >>> pair_marg = sampleset.info['interaction_marginals']
        >>> round(pair_marg[('a', 'b')][(1, -1)], 3)  # prob(a == 1 & b == -1)
        0.001

    .. _tree decomposition: https://en.wikipedia.org/wiki/Tree_decomposition

    .. _Boltzmann distribution: https://en.wikipedia.org/wiki/Boltzmann_distribution

    """
    parameters = {'num_reads': [],
                  'elimination_order': ['max_treewidth'],
                  'beta': [],
                  'marginals': [],
                  'seed': []}
    """Keyword arguments accepted by the sampling methods.

    Examples:

        >>> from dwave.samplers import TreeDecompositionSampler
        >>> sampler = TreeDecompositionSampler()
        >>> sampler.parameters.keys()
        dict_keys(['num_reads', 'elimination_order', 'beta', 'marginals', 'seed'])

    See :meth:`.sample` for descriptions.

    """

    properties = {'max_treewidth': 25}
    """Information about the sampler.

    Examples:

        >>> from dwave.samplers import TreeDecompositionSampler
        >>> sampler = TreeDecompositionSampler()
        >>> sampler.properties
        {'max_treewidth': 25}

    """

    def __init__(self):
        # make these object properties rather than class properties
        self.parameters = dict(self.parameters)
        self.properties = dict(self.properties)

    def sample(self, bqm: dimod.BinaryQuadraticModel, num_reads: Optional[int] = 1,
               elimination_order: Optional[List[Variable]] = None, beta: Optional[float] = 3.0,
               marginals: Optional[bool] = True, seed: Optional[int] = None) -> dimod.SampleSet:
        """Draw samples and compute marginals of a binary quadratic model.

        Args:
            bqm:
                Binary quadratic model.

            num_reads:
                Number of samples to draw.

            elimination_order:
                Variable elimination order. Should be a list of the
                variables in the binary quadratic model. If None, the min-fill
                heuristic [#gd]_ is used to generate one.

            beta:
                `Boltzmann distribution`_ inverse temperature parameter.

            marginals:
                Whether or not to compute the marginals. If True, they are
                included in the return :obj:`~dimod.SampleSet`'s ``info`` field.
                See example in :class:`.TreeDecompositionSampler`.

            seed:
                Random number generator seed. Negative values cause a
                time-based seed to be used.

        Returns:
            Returned :attr:`dimod.SampleSet.info` contains:

                * ``'log_partition_function'``: The log partition function.


            If ``marginals=True``, also contains:

                * ``'variable_marginals'``: Dict of the form ``{v: p, ...}``, where
                  ``v`` is a variable in the binary quadratic model and
                  ``p = prob(v == 1)``.
                * ``'interaction_marginals'``: Dict of the form
                  ``{(u, v): {(s, t): p, ...}, ...}``, where ``(u, v)`` is an
                  interaction in the binary quadratic model and
                  ``p = prob(u == s & v == t)``.

        Raises:
            ValueError:
                The treewidth_ of the given bqm and elimination order cannot
                exceed the value provided in :attr:`.properties`.

        .. _treewidth: https://en.wikipedia.org/wiki/Treewidth
        .. _Boltzmann distribution: https://en.wikipedia.org/wiki/Boltzmann_distribution
        .. [#gd] Gogate & Dechter, "A Complete Anytime Algorithm for Treewidth",
           https://arxiv.org/abs/1207.4109

        """
        if not bqm:
            info = {'log_partition_function': 0.0}
            if marginals:
                info['variable_marginals'] = {}
                info['interaction_marginals'] = {}
            samples = np.empty((num_reads, 0), dtype=samples_dtype)
            energies = bqm.energies(samples, dtype=energies_dtype)
            return dimod.SampleSet.from_samples(samples, bqm.vartype,
                                                energy=energies, info=info)

        bqm = dimod.as_bqm(bqm, copy=True, dtype=float)

        if elimination_order is None:
            # note that this does not respect the given seed
            tree_width, elimination_order = min_fill_heuristic(bqm)
        else:
            # this also checks the order against the bqm
            tree_width = elimination_order_width(bqm, elimination_order)

        # developer note: we start getting bad_alloc errors above tree_width 25, this
        # should be fixed in the future
        if tree_width > self.properties['max_treewidth']:
            raise ValueError(
                f"maximum treewidth of {self.properties['max_treewidth']} exceeded. "
                "To see the bqm's treewidth:\n"
                ">>> import dwave_networkx as dnx\n"
                f">>> dnx.elimination_order_width(bqm.adj, {elimination_order})"
                )

        # relabel bqm variables so that we only work with linear indices
        bqm_copy, int_to_var = bqm.relabel_variables_as_integers(inplace=False)
        var_to_int = {v: k for k, v in int_to_var.items()}

        # relabel variables in the elimination order as well
        elimination_order = [var_to_int.get(var, var) for var in elimination_order]

        max_complexity = tree_width + 1

        samples, data = sample_bqm_wrapper(bqm=bqm_copy,
                                           beta=beta,
                                           max_complexity=max_complexity,
                                           order=elimination_order,
                                           marginals=marginals,
                                           num_reads=num_reads,
                                           seed=seed)

        info = {'log_partition_function': data['log_partition_function']}

        if marginals:
            info['variable_marginals'] = {}
            for i in elimination_order:
                info['variable_marginals'][int_to_var.get(i, i)] = data['variable_marginals'][i]

            info['interaction_marginals'] = {}
            low = -1 if bqm.vartype is dimod.SPIN else 0
            configs = (low, low), (1, low), (low, 1), (1, 1)
            for (i, j), probs in zip(data['interactions'],
                                     data['interaction_marginals']):
                u = int_to_var.get(i, i)
                v = int_to_var.get(j, j)
                info['interaction_marginals'][(u, v)] = dict(zip(configs, probs))

        energies = bqm.energies((samples, bqm.variables),
                                dtype=energies_dtype)

        return dimod.SampleSet.from_samples((samples, bqm.variables),
                                            bqm.vartype,
                                            energy=energies,
                                            info=info)



# Copyright 2019 D-Wave Systems Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""A dimod sampler that uses the steepest gradient descent."""

from numbers import Integral
from typing import Optional

from dimod.core.initialized import InitialStateGenerator

import dimod
import numpy as np

from dwave.samplers.greedy.descent import steepest_gradient_descent

__all__ = ["SteepestDescentSolver", "SteepestDescentSampler"]


class SteepestDescentSolver(dimod.Sampler, dimod.Initialized):
    """Steepest descent sampler for binary quadratic models.

    `Steepest descent <https://en.wikipedia.org/wiki/Gradient_descent>`_  is the
    discrete analogue of gradient descent, but the best move is computed using a
    local minimization rather rather than computing a gradient. The dimension along
    which to descend is determined, at each step, by the variable flip that causes
    the greatest reduction in energy.

    Solves convex problems to optimality.

    Number of downhill runs (samples produced) is determined by ``num_reads``,
    number of ``initial_states``, or a combination of the two, depending on
    the ``initial_states_generator``.

    For a given input model's graph :math:`G = (V, E)`, :math:`V` being a set of
    graph vertices and :math:`E` a set of edges, runtime complexity of the
    underlying C++ implementation is :math:`O(|E|)` for initialization phase
    and :math:`O(|V|)` per downhill step.

    In the ``large_sparse_opt`` mode, runtime complexity on sparse graphs is
    :math:`O(|V|*log|V|)` for initialization and :math:`O(max\_degree * log|V|)`
    per downhill step.

    Aliased as :class:`~greedy.sampler.SteepestDescentSampler`.

    Examples:
        Solve a simple Ising problem.

        >>> from dwave.samplers import SteepestDescentSampler
        ...
        >>> sampler = SteepestDescentSampler()
        >>> samples = sampler.sample_ising({0: 2, 1: 2}, {(0, 1): -1})
        ...
        >>> print(samples)      # doctest: +SKIP
           0  1 energy num_oc. num_st.
        0 -1 -1   -5.0       1       2
        ['SPIN', 1 rows, 1 samples, 2 variables]

        Post-processes samples generated by another sampler (simulated annealing
        in this example):

        >>> from dwave.samplers import SteepestDescentSampler, SimulatedAnnealingSampler
        >>> import dimod
        ...
        >>> bqm = dimod.generators.ran_r(5, 3)
        >>> samples = SimulatedAnnealingSampler().sample(bqm)
        >>> postprocessed = SteepestDescentSampler().sample(bqm, initial_states=samples)

        For additional examples, see :meth:`.sample`.
    """

    parameters = None
    """Keyword arguments accepted by the sampling methods.

    Examples:

        >>> from dwave.samplers import SteepestDescentSampler
        >>> sampler = SteepestDescentSampler()
        >>> sampler.parameters.keys()
        dict_keys(['num_reads', 'initial_states', 'initial_states_generator', 'seed', 'large_sparse_opt'])

    """

    properties = None
    """Values for parameters accepted by the sampling methods.

    Examples:

        >>> from dwave.samplers import SteepestDescentSampler
        >>> sampler = SteepestDescentSampler()
        >>> sampler.properties.keys()
        dict_keys(['initial_states_generators', 'large_sparse_opt_values'])

    """

    def __init__(self):
        # create a copy (isolate from subclass)
        self.parameters = {
            'num_reads': [],
            'initial_states': [],
            'initial_states_generator': ['initial_states_generators'],
            'seed': [],
            'large_sparse_opt': ['large_sparse_opt_values'],
        }
        self.properties = {
            'initial_states_generators': ('none', 'tile', 'random'),
            'large_sparse_opt_values': (True, False),
        }

    def sample(self, bqm: dimod.BinaryQuadraticModel,
               num_reads: Optional[int] = None,
               initial_states: Optional[dimod.typing.SamplesLike] = None,
               initial_states_generator: InitialStateGenerator = "random",
               seed: Optional[int] = None,
               large_sparse_opt: bool = False, **kwargs) -> dimod.SampleSet:
        """Sample from a binary quadratic model.

        Starts from ``initial_states``, and converges to local minima using
        discrete steepest-descent method.

        Args:
            bqm: Binary quadratic model to be sampled.

            num_reads:
                Number of reads. Each read is generated by one run of the steepest
                descent algorithm. If ``num_reads`` is not explicitly given, it is
                selected to match the number of initial states given. If initial states
                are not provided, only one read is performed.

            initial_states:
                One or more samples, each defining an initial state for all the
                problem variables. Initial states are given one per read, but
                if fewer than ``num_reads`` initial states are defined, additional
                values are generated as specified by ``initial_states_generator``.
                See :func:`~dimod.as_samples` for a description of "samples-like".

            initial_states_generator:
                Defines the expansion of ``initial_states`` if fewer than
                ``num_reads`` are specified:

                * "none":
                    If the number of initial states specified is smaller than
                    ``num_reads``, raises ``ValueError``.

                * "tile":
                    Reuses the specified initial states if fewer than ``num_reads``
                    or truncates if greater.

                * "random":
                    Expands the specified initial states with randomly generated
                    states if fewer than ``num_reads`` or truncates if greater.

            seed:
                32-bit unsigned integer seed to use for the PRNG. Specifying a
                particular seed with a constant set of parameters produces
                identical results. If not provided, a random seed is chosen.

            large_sparse_opt:
                Use optimizations for large and sparse problems (search tree for
                next descent variable instead of linear array).

        Note:
            Number of descends (single variable flips) taken to reach the local
            minimum for each sample is stored in a data vector called ``num_steps``.

        Examples:
            This example samples a simple two-variable Ising model.

            >>> import dimod
            >>> bqm = dimod.BQM.from_ising({}, {'ab': 1})
            ...
            >>> from dwave.samplers import SteepestDescentSampler
            >>> sampler = SteepestDescentSampler()
            ...
            >>> samples = sampler.sample(bqm)
            >>> samples.first.energy
            -1.0

            Run steepest descent one million times (takes ~150ms on an average
            laptop), converging to local minima, each time starting from a
            random state:

            >>> bqm = dimod.BQM.from_ising({}, {'ab': 1})
            >>> sampler = SteepestDescentSampler()
            ...
            >>> samples = sampler.sample(bqm, num_reads=10**6)
            >>> print(samples.aggregate())      # doctest: +SKIP
               a  b energy num_oc. num_st.
            0 -1 +1   -1.0  500115       1
            1 +1 -1   -1.0  499885       1
            ['SPIN', 2 rows, 1000000 samples, 2 variables]

            Use a combination of one fixed initial state and two randomly
            generated ones to sample from a simple convex Ising problem with
            global minimum at ``(-1,-1)``:

            >>> bqm = dimod.BQM.from_ising({'x': 2, 'y': 2}, {'xy': -1})
            >>> sampler = SteepestDescentSampler()
            ...
            >>> samples = sampler.sample(bqm, initial_states=([1, 1], 'xy'), num_reads=3)
            >>> print(samples)      # doctest: +SKIP
               x  y energy num_oc. num_st.
            0 -1 -1   -5.0       1       2
            1 -1 -1   -5.0       1       1
            2 -1 -1   -5.0       1       2
            ['SPIN', 3 rows, 3 samples, 2 variables]

            Notice it required 2 variable flips (``num_steps`` field in the last
            column) to reach the minimum state, ``(-1, -1)``, from the initial
            state, ``(1, 1)``.
        """

        # get the original vartype so we can return consistently
        original_vartype = bqm.vartype

        # convert to spin
        if bqm.vartype is not dimod.SPIN:
            bqm = bqm.change_vartype(dimod.SPIN, inplace=False)

        # validate seed
        if not (seed is None or isinstance(seed, Integral)):
            raise TypeError("'seed' should be None or a positive 32-bit integer")
        if isinstance(seed, Integral) and not 0 <= seed <= 2**32 - 1:
            raise ValueError("'seed' should be an integer between 0 and 2**32 - 1 inclusive")

        # parse initial_states et al
        parsed_initial_states = self.parse_initial_states(
            bqm,
            num_reads=num_reads,
            initial_states=initial_states,
            initial_states_generator=initial_states_generator,
            seed=seed)

        num_reads = parsed_initial_states.num_reads
        initial_states = parsed_initial_states.initial_states

        # get linear/quadratic data
        linear, (coupler_starts, coupler_ends, coupler_weights), offset = \
            bqm.to_numpy_vectors(variable_order=initial_states.variables)

        # we need initial states as contiguous numpy array
        initial_states_array = \
            np.ascontiguousarray(initial_states.record.sample, dtype=np.int8)

        # run the steepest descent
        samples, energies, num_steps = steepest_gradient_descent(
            num_reads,
            linear, coupler_starts, coupler_ends, coupler_weights,
            initial_states_array, large_sparse_opt)

        # resulting sampleset
        result = dimod.SampleSet.from_samples(
            (samples, initial_states.variables),
            energy=energies + offset,
            vartype=dimod.SPIN,
            num_steps=num_steps,
        )

        result.change_vartype(original_vartype, inplace=True)

        return result


SteepestDescentSampler = SteepestDescentSolver



# Copyright 2022 D-Wave Systems Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from setuptools import setup
from Cython.Build import cythonize
from setuptools.command.build_ext import build_ext

import dimod
import numpy


class build_ext_with_args(build_ext):
    """Add compiler-specific compile/link flags."""

    extra_compile_args = {
        'msvc': ['/std:c++14'],
        'unix': ['-std=c++11'],
    }

    extra_link_args = {
        'msvc': [],
        'unix': ['-std=c++11'],
    }

    def build_extensions(self):
        compiler = self.compiler.compiler_type

        compile_args = self.extra_compile_args[compiler]
        for ext in self.extensions:
            ext.extra_compile_args = compile_args

        link_args = self.extra_link_args[compiler]
        for ext in self.extensions:
            ext.extra_compile_args = link_args

        super().build_extensions()



setup(
    cmdclass={'build_ext': build_ext_with_args},
    ext_modules=cythonize(
        ['dwave/samplers/greedy/descent.pyx',
         'dwave/samplers/random/*.pyx',
         'dwave/samplers/sa/*.pyx',
         'dwave/samplers/tabu/tabu_search.pyx',
         'dwave/samplers/tree/*.pyx',
         ],
        ),
    include_dirs=[
        dimod.get_include(),
        numpy.get_include(),
        ],

    )
